<?php

  declare( strict_types = 1 );

  namespace Laodc\Dns;

  use Laodc\Database\Database;
  use Laodc\Errors\Errors;
  use Laodc\Functions\Functions;

  class Domain extends Base
  {
    private Database $db;
    private Records $records;
    private array $options = [];

    function __construct( Database $database, array $domain )
    {
      $this->db = $database;

      $this->populate( [ Validate::class, 'formatDomainData' ], $domain );

      $this->records = new Records( $this->db, $this );

      if( isset( $domain['id'] ) )
        $this->options = ( new Options( $this->db, ( int )$domain['id'] ) )->toArray();

      $default_options = [
        'recursive'       => 'FALSE',
        'dnssec'          => 'FALSE',
        'autosync'        => 'FALSE',
        'generate_ptr'    => 'FALSE',
        'generate_a'      => 'FALSE',
        'generate_aaaa'   => 'FALSE',
      ];

      $this->options = array_merge( $default_options, $this->options );
    }

    /**
     * get record class for the domain
     *
     * @param bool $asArray
     *        [optional] output records as an array
     *
     * @return Records|array
     *        the associated Records object or array
     */
    public function Records( bool $asArray = false ) : Records|array
    {
      if( false === $asArray )
        return $this->records;
      
      // prepare array
      $records = [];
      foreach( $this->records->get( 'records' ) as $record )
        $records[] = $record->toArray();

      return $records;
    }

    public function recordById( int $id ) :? Record
    {
      return self::Records()->getById( $id );
    }

    /**
     * grab list of existing options for the domain
     *
     * @return array
     *        array of available options
     */
    public function options() : array
    {
      return $this->options;
    }

    /**
     * option to enable or disable recursive for the domain
     *
     * @param bool $status
     *        enable or disable status
     *
     * @return self
     */
    public function setRecursive( bool $status ) : self
    {
      $this->options['recursive'] = Validate::boolToSQL( $status );

      return $this;
    }

    /**
     * option to enable or disable dnssec for the domain
     *
     * @param bool $status
     *        enable or disable status
     *
     * @return self
     */
    public function setDnssec( bool $status ) : self
    {
      $this->options['dnssec'] = Validate::boolToSQL( $status );
      
      return $this;
    }

    /**
     * option to enable or disable auto sync for the domain
     *
     * @param bool $status
     *        enable or disable status
     *
     * @return self
     */
    public function setAutoSync( bool $status ) : self
    {
      $this->options['autosync'] = Validate::boolToSQL( $status );
      
      return $this;
    }

    /**
     * option to enable or disable generate PTR for the domain
     *
     * @param bool $status
     *        enable or disable status
     *
     * @return self
     */
    public function setGeneratePtr( bool $status ) : self
    {
      $this->options['generate_ptr'] = Validate::boolToSQL( $status );
      
      return $this;
    }

    /**
     * option to enable or disable generate A for the domain
     *
     * @param bool $status
     *        enable or disable status
     *
     * @return self
     */
    public function setGenerateA( bool $status ) : self
    {
      $this->options['generate_a'] = Validate::boolToSQL( $status );

      return $this;
    }

    /**
     * option to enable or disable generate AAAA for the domain
     *
     * @param bool $status
     *        enable or disable status
     *
     * @return self
     */
    public function setGenerateAAAA( bool $status ) : self
    {
      $this->options['generate_aaaa'] = Validate::boolToSQL( $status );

      return $this;
    }

    /**
     * the IPv4 PTR string formatting that will be use for BIND output
     *
     * @param string $format
     *        the BIND formatted string
     *
     * @return self
     */
    public function setGeneratePtrFormat( string $format ) : self
    {
      // validate string format
      if( null === Validate::Name( $format, null, false, false, '$' ) )
        Errors::set( 'OPTION_PTR_FORMAT: ' . Errors::get()[0]['error'] );
      else
        $this->options['ptr_format'] = $format;
      
      return $this;
    }

    /**
     * the A string formatting that will be use for BIND output
     *
     * @param string $format
     *        the BIND formatted string
     *
     * @return self
     */
    public function setGenerateAFormat( string $format ) : self
    {
      // validate string format
      if( null === Validate::Name( $format, null, false, false, '$' ) )
        Errors::set( 'OPTION_A_FORMAT: ' . Errors::get()[0]['error'] );
      else
        $this->options['a_format'] = $format;
      
      return $this;
    }

    /**
     * the AAAA string formatting that will be use for BIND output
     *
     * @param string $format
     *        the BIND formatted string
     *
     * @return self
     */
    public function setGenerateAAAAFormat( string $format ) : self
    {
      // validate string format
      if( null === Validate::Name( $format, null, false, false, '$:' ) )
        Errors::set( 'OPTION_AAAA_FORMAT: ' . Errors::get()[0]['error'] );
      else
        $this->options['aaaa_format'] = $format;
      
      return $this;
    }

    /**
     * save the domain information to database
     *
     * @return bool
     *        true is saving was successful
     */
    public function save() : bool
    {
      // allowed fields
      $fields = [
        'fqdn'      => true,
        'ns'        => true,
        'mbox'      => true,
        'refresh'   => false,
        'retry'     => false,
        'expire'    => false,
        'minimum'   => false,
        'ttl'       => false,
        'xfer'      => false,
        'active'    => false,
        'owner_id'  => true,
      ];

      foreach( $fields as $field => $required )
      {
        if( true === $required && empty( $this->get( $field ) ) )
        {
          Errors::set( sprintf( '%s: Required field', strtoupper( $field ) ) );
          return false;
        }

        // sanatize record data
        $this->set( $field, Validate::formatDomainData( $field, $this->get( $field ) ) );
      }

      // make sure fqdn is valid
      if( null === ( $fqdn = Validate::Name( $this->get( 'fqdn' ) ) ) )
      {
        Errors::set( 'FQDN: ' . Errors::get()[0]['error'] );
        return false;
      }

      // make sure it ends in a dot
      if( '.' != substr( $fqdn, -1 ) )
        $fqdn .= '.';

      $this->set( 'fqdn', $fqdn );

      // make sure ns is valid
      if( null === ( $ns = Validate::Name( $this->get( 'ns' ) ) ) )
      {
        Errors::set( 'NS: ' . Errors::get()[0]['error'] );
        return false;
      }

      // make sure it ends in a dot
      if( '.' != substr( $ns, -1 ) )
        $ns .= '.';

      $this->set( 'ns', $ns );

      // make sure mbox is valid
      if( null === ( $mbox = Validate::Name( $this->get( 'mbox' ) ) ) )
      {
        Errors::set( 'MBOX: ' . Errors::get()[0]['error'] );
        return false;
      }

      // make sure it ends in a dot
      if( '.' != substr( $mbox, -1 ) )
        $mbox .= '.';

      $this->set( 'mbox', $mbox );

      // check that serial is correct format
      $this->set( 'serial', Validate::generateSerial( $this->get( 'serial' ) ) );

      if( false === Validate::isInt64( $this->get( 'serial' ) ) )
      {
        Errors::set( 'SERIAL: ' . Errors::get()[0]['error'] );
        return false;
      }

      // check that refresh is correct format
      $this->set( 'refresh', $this->get( 'refresh' ) ?? Domains::DEFAULT_REFRESH );

      if( false === Validate::isInt32( $this->get( 'refresh' ), Domains::MIN_REFRESH ) )
      {
        Errors::set( 'REFRESH: ' . Errors::get()[0]['error'] );
        return false;
      }

      // check that retry is correct format
      $this->set( 'retry', $this->get( 'retry' ) ?? Domains::DEFAULT_RETRY );

      if( false === Validate::isInt32( $this->get( 'retry' ), Domains::MIN_RETRY ) )
      {
        Errors::set( 'RETRY: ' . Errors::get()[0]['error'] );
        return false;
      }

      // check that expire is correct format
      $this->set( 'expire', $this->get( 'expire' ) ?? Domains::DEFAULT_EXPIRE );

      if( false === Validate::isInt32( $this->get( 'expire' ), Domains::MIN_EXPIRE ) )
      {
        Errors::set( 'EXPIRE: ' . Errors::get()[0]['error'] );
        return false;
      }

      // check that minimum is correct format
      $this->set( 'minimum', $this->get( 'minimum' ) ?? Domains::DEFAULT_TTL );

      if( false === Validate::isInt32( $this->get( 'minimum' ), Domains::MIN_TTL ) )
      {
        Errors::set( 'MINIMUM: ' . Errors::get()[0]['error'] );
        return false;
      }

      // check that ttl is correct format
      $this->set( 'ttl', $this->get( 'ttl' ) ?? Domains::DEFAULT_TTL );

      if( false === Validate::isInt32( $this->get( 'ttl' ), Domains::MIN_TTL ) )
      {
        Errors::set( 'TTL: ' . Errors::get()[0]['error'] );
        return false;
      }

      // check that xfer is list of valid ip addresses
      if( null === ( $xfer = Validate::XFER( $this->get( 'xfer' ) ) ) )
      {
        Errors::set( 'XFER: ' . Errors::get()[0]['error'] );
        return false;
      }

      $this->set( 'xfer', $xfer );

      // check for existing domain
      // if ID is specified, don't match against itself
      $sql = sprintf( '
        SELECT
          id
        FROM
          %s
        WHERE
          %s
          fqdn LIKE \'%s\'
        ',
        $this->db->escape( Records::TABLE_NAME ),
        ( $this->get( 'id' ) ) ? sprintf( 'id <> %d AND', $this->get( 'id' ) ) : '',
        $this->db->escape( $this->get( 'fqdn' ) )
      );

      // domain exists, bail out
      if( [] !== $this->db->fetch( $sql ) )
      {
        Errors::set( 'FQDN: Existing domain found' );
        return false;
      }

      // add new domain
      $sql = sprintf( '
        INSERT INTO
          %s
        (
          owner_id, fqdn, ns, mbox, serial,
          refresh, retry, expire, minimum, ttl,
          active, xfer
        )
        VALUES
          ( %d, \'%s\', \'%s\', \'%s\', %d, %d, %d, %d, %d, %d, %s, \'%s\' )
        RETURNING
          id, owner_id, fqdn, ns, mbox, serial,
          refresh, retry, expire, minimum, ttl,
          active, xfer
        ',
        $this->db->escape( Domains::TABLE_NAME ),
        $this->get( 'owner_id' ),
        $this->db->escape( $this->get( 'fqdn' ) ),
        $this->db->escape( $this->get( 'ns' ) ),
        $this->db->escape( $this->get( 'mbox' ) ),
        $this->get( 'serial' ),
        $this->get( 'refresh' ),
        $this->get( 'retry' ),
        $this->get( 'expire' ),
        $this->get( 'minimum' ),
        $this->get( 'ttl' ),
        Validate::boolToSQL( $this->get( 'active' ) ),
        $this->db->escape( $this->get( 'xfer' ) )
      );

      // update domain if ID exists
      if( !empty( $this->get( 'id' ) ) )
      {
        $sql = sprintf( '
          UPDATE
            %s
          SET
            owner_id = %d,
            fqdn = \'%s\',
            ns = \'%s\',
            mbox = \'%s\',
            serial = %d,
            refresh = %d,
            retry = %d,
            expire = %d,
            minimum = %d,
            ttl = %d,
            active = %s,
            xfer = \'%s\'
          WHERE
            id = %d
          RETURNING
            id, owner_id, fqdn, ns, mbox, serial,
            refresh, retry, expire, minimum, ttl,
            active, xfer
          ',
          $this->db->escape( Domains::TABLE_NAME ),
          $this->get( 'owner_id' ),
          $this->db->escape( $this->get( 'fqdn' ) ),
          $this->db->escape( $this->get( 'ns' ) ),
          $this->db->escape( $this->get( 'mbox' ) ),
          $this->get( 'serial' ),
          $this->get( 'refresh' ),
          $this->get( 'retry' ),
          $this->get( 'expire' ),
          $this->get( 'minimum' ),
          $this->get( 'ttl' ),
          Validate::boolToSQL( $this->get( 'active' ) ),
          $this->db->escape( $this->get( 'xfer' ) ),
          $this->get( 'id' )
        );
      }

      if( [] === ( $domain = $this->db->fetch( $sql ) ) )
      {
        Errors::set( 'FQDN: Unable to add domain to database' );
        return false;
      }

      $domain['active'] = Validate::toBool( $domain['active'] );

      // configure options for domain
      $options = new Options( $this->db, ( int )$domain['id'] );

      foreach( $this->options as $k => $v )
        $options->set( $k, $v );

      // set up fields for domain
      foreach( $domain as $field => $value )
        $this->set( $field, $value );

      return true;
    }

    /**
     * remove domain from database
     *
     * @return bool
     *        true if successfully removed
     */
    public function del() : bool
    {
      if( empty( $this->get( 'id' ) ) )
      {
        Errors::set( 'FQDN: Domain does not exist' );
        return false;
      }

      // delete record
      $sql = sprintf( '
        DELETE FROM
          %s
        WHERE
          id = %d
        ',
        Domains::TABLE_NAME,
        $this->get( 'id' )
      );

      if( false === $this->db->query( $sql ) )
      {
        Errors::set( 'FQDN: Error deleting domain' );
        return false;
      }

      ( new Options( $this->db, ( int )$this->get('id' ) ) )->del_all();

      // reset fields
      $this->data_fields = [];
      return true;
    }

    /**
     * update domain with new serial
     *
     * @return bool
     *        true on successful update
     */
    public function updateSerial() : bool
    {
      $serial = Validate::generateSerial( intval( $this->get( 'serial' ) ) );

      $sql = sprintf( '
        UPDATE
          %s
        SET
          serial = %d
        WHERE
          id = %d
        RETURNING
          serial
        ',
        Domains::TABLE_NAME,
        $serial,
        $this->get( 'id' )
      );

      if( [] === ( $record = $this->db->fetch( $sql ) ) )
        return false;

      // update serial in object
      $this->set( 'serial', $record['serial'] );

      return true;
    }

    /**
     * export domain to output folder
     * 
     * @param string $outputFolder
     *        path of where to output the BIND zone file
     * 
     * @return bool
     *        true on successful export
     */
    public function export( string $outputFolder ) : bool
    {
      if( empty( $this->get( 'id' ) ) )
      {
        Errors::set( 'FQDN: Invalid domain to export' );

        // clean up files if domain failed to export
        Functions::removeFolder( $outputFolder );

        return false;
      }

      $date = date( 'c' );
      $zone = substr( $this->get( 'fqdn' ), 0, -1 );

      $serial = str_pad( $this->get( 'serial' ), 15, ' ', STR_PAD_RIGHT );
      $refresh = str_pad( $this->get( 'refresh' ), 15, ' ', STR_PAD_RIGHT );
      $retry = str_pad( $this->get( 'retry' ), 15, ' ', STR_PAD_RIGHT );
      $expire = str_pad( $this->get( 'expire' ), 15, ' ', STR_PAD_RIGHT );
      $minimum = str_pad( $this->get( 'minimum' ), 15, ' ', STR_PAD_RIGHT );

      $output = <<< DNS
\$TTL {$this->get( 'ttl' )}
; Zone: {$this->get( 'fqdn' )} (#{$this->get( 'id' )})
; {$date}
\$ORIGIN {$this->get( 'fqdn' )}

@\tIN\tSOA\t{$this->get( 'ns' )}\t{$this->get( 'mbox' )}  (
\t\t{$serial}; Serial
\t\t{$refresh}; Refresh
\t\t{$retry}; Retry
\t\t{$expire}; Expire
\t\t{$minimum}) ; Minimum


DNS;

      if( [] === ( $records = $this->Records()->getAll() ) )
      {
        Errors::set( 'FQDN: No records for domain' );

        // clean up files if domain failed to export
        Functions::removeFolder( $outputFolder );

        return false;
      }

      $max = 0;
      foreach( $records as $record )
      {
        if( true === $record->get( 'active' ) )
          $max = ( strlen( $record->get( 'name' ) ) > $max ) ? strlen( $record->get( 'name' ) ) : $max;
      }

      if( 0 === $max )
      {
        Errors::set( 'FQDN: No records to export' );

        // clean up files if domain failed to export
        Functions::removeFolder( $outputFolder );

        return false;
      }

      $max = ( ( $max < 6 ) ? 6 : $max ) + 3; // min max + buffer

      $str = [
        'name' => str_pad( ';name', $max, ' ', STR_PAD_RIGHT ),
        'ttl' => str_pad( 'ttl', 6, ' ', STR_PAD_RIGHT ),
        'type' => str_pad( 'type', 6, ' ', STR_PAD_RIGHT ),
        'pref' => str_pad( 'pref', 5, ' ', STR_PAD_RIGHT ),
        'data' => 'data',
      ];


      $output .= <<< DNS
{$str['name']}{$str['ttl']}   {$str['type']}{$str['pref']}{$str['data']}

DNS;

      $append = [];
      foreach( $records as $record )
      {
        if( false === $record->get( 'active' ) )
          continue;

        if( '*' == substr( $record->get( 'name' ), 0, 1 ) )
        {
          $append[] = $record;
          continue;
        }

        $output .= self::exportRecord( $record, $max );
      }

      foreach( $append as $record )
      {
        $output .= self::exportRecord( $record, $max );
      }

      $file = sprintf( '%s/dns.%s', $outputFolder, $zone );
      file_put_contents( $file, $output );

      return true;
    }

    /**
     * export of single domain record
     * 
     * @param Record $record
     *        the record to format
     * 
     * @param int $length
     *        padding length for visual alignment
     * 
     * @return string
     *        the formatted BIND line for the record
     */
    private function exportRecord( Record $record, int $length ) : string
    {
      $name = str_pad( $record->get( 'name' ), $length, ' ', STR_PAD_RIGHT );
      $ttl = str_pad( $record->get( 'ttl' ), 6, ' ', STR_PAD_RIGHT );
      $type = str_pad( $record->get( 'type' ), 6, ' ', STR_PAD_RIGHT );
      $pref = str_pad( $record->get( 'pref' ), 5, ' ', STR_PAD_RIGHT );

      $output = '';
      switch( strtoupper( $record->get( 'type' ) ) )
      {
        case Record::TYPE_MX:
          $output = <<< DNS
{$name}{$ttl}IN {$type}{$pref}{$record->get( 'data' )}

DNS;
          break;
        case Record::TYPE_SRV:
        case Record::TYPE_TLSA:
          list( $weight, $port, $record ) = explode( ' ', $record->get( 'data' ) );
          $weight = str_pad( $weight, 5, ' ', STR_PAD_RIGHT );
          $port = str_pad( $port, 5, ' ', STR_PAD_RIGHT );

          $output = <<< DNS
{$name}{$ttl}IN {$type}{$pref}{$weight}{$port}{$record}

DNS;
          break;
        case Record::TYPE_CAA:
          list( $left, $right ) = explode( ' ', $record->get( 'data' ) );

          $left = str_pad( $left, 10, ' ', STR_PAD_RIGHT );

          $output = <<< DNS
{$name}{$ttl}IN {$type}     {$left}{$right}

DNS;
          break;
        case Record::TYPE_TXT:
          $record = '"' . str_replace( [ "\0", '"' ], [ '" "', "\\\"" ], $record->get( 'data' ) ) . '"';

          $output = <<< DNS
{$name}{$ttl}IN {$type}     {$record}

DNS;
          break;
        case Record::TYPE_URL:
        case Record::TYPE_URL301:

          // URL redirectors will go through redirect.laodc.com CNAME record
          $type = str_pad( 'CNAME', 6, ' ', STR_PAD_RIGHT );
          $address = Domains::URL_REDIRECT_CNAME;

          // if name is the domain fqdn we have to use IP address instead of CNAME
          if( $record->get( 'name' ) == $this->get( 'fqdn' ) )
          {
            $type = str_pad( 'A', 6, ' ', STR_PAD_RIGHT );
            $address = Domains::URL_REDIRECT_A;
          }

          $output = <<< DNS
{$name}{$ttl}IN {$type}     {$address}

DNS;

          break;
        default:
          $pref = str_pad( ' ', 5, ' ', STR_PAD_RIGHT );
          $output = <<< DNS
{$name}{$ttl}IN {$type}     {$record->get( 'data' )}

DNS;
          break;
      }

      return $output;
    }
  }
