<?php

  namespace Laodc\Dns;

  use Laodc\Database\Database;
  use Laodc\Errors\Errors;

  class Domains
  {
    // Main table name
    const TABLE_NAME = 'domains';

    // Auto settings
    const AUTO_UPDATE_SERIALS = TRUE;
    const AUTO_UPDATE_PTRS = FALSE;
    const AUTO_DEFAULTS = TRUE;

    // TTL Settings
    const MIN_TTL = 60;
    const MIN_REFRESH = 300;
    const MIN_RETRY = 300;
    const MIN_EXPIRE = 86400;

    const DEFAULT_TTL = 1800;
    const DEFAULT_REFRESH = 1800;
    const DEFAULT_RETRY = 1800;
    const DEFAULT_EXPIRE = 86400;
    const DEFAULT_MIN_TTL = 1800;

    // Master zones
    const MASTER_TABLE_NAME = 'masters';
    const EXPORT_MASTER = 'master';
    const EXPORT_SLAVE = 'slave';
    const EXPORT_ALL = -1;

    // URI Redirct domain
    // @TODO
    // Grab this from database instead of hardcoded
    const URL_REDIRECT_A = '103.26.103.36';
    const URL_REDIRECT_CNAME = 'redirect.laodc.com.';

    private int $group_size = 0;

    private Database $db;

    function __construct( Database $database, string $fqdn = null )
    {
      $this->db = $database;

      if( !empty( $fqdn ) )
        $this->getByName( $fqdn );
    }

    public static function toArray( array $array ) : array
    {
      $domains = [];

      foreach( $array as $domain )
        $domains[] = $domain->toArray();

      return $domains;
    }

    /**
     * get a list of all domains in the database
     * 
     * @return array
     *        array of domains
     */
    public function getAll() : array
    {
      $sql = sprintf( '
        SELECT
          id, fqdn, ns, mbox, serial,
          refresh, retry, expire, minimum, ttl,
          active, xfer
        FROM
          %s
        ORDER BY
          fqdn
        ',
        $this->db->escape( self::TABLE_NAME )
      );

      $domains = [];

      foreach( $this->db->fetchAll( $sql ) as $domain )
      {
        $domain['active'] = Validate::toBool( $domain['active'] );
        $domains[$domain['id']] = new Domain( $this->db, $domain );
      }

      return $domains;
    }

    /**
     * get domain based on it's database ID
     * 
     * @param int $id
     *        domain ID in database
     * 
     * @return Domain|null
     *        returns matching domain or null if not found
     */
    public function getById( int $id ) :? Domain
    {
      $sql = sprintf( '
        SELECT
          id, fqdn, ns, mbox, serial,
          refresh, retry, expire, minimum, ttl,
          active, xfer, owner_id
        FROM
          %s
        WHERE
          id = %d
        LIMIT
          1
        ',
        $this->db->escape( self::TABLE_NAME ),
        $id
      );

      if( [] === ( $result = $this->db->fetch( $sql ) ) )
      {
        Errors::set( 'Domain not found' );
        return null;
      }

      $result['active'] = Validate::toBool( $result['active'] );

      return new Domain( $this->db, $result );
    }

    /**
     * get domain based on FQDN address
     * 
     * @param string $fqdn
     *        the fqdn of the domain record to obtain
     * 
     * @return Domain|null
     *        returns the matching domain or null if not found
     */
    public function getByName( string $fqdn ) :? Domain
    {
      $sql = sprintf( '
        SELECT
          id, fqdn, ns, mbox, serial,
          refresh, retry, expire, minimum, ttl,
          active, xfer, owner_id
        FROM
          %s
        WHERE
          fqdn LIKE \'%s\'
        LIMIT
          1
        ',
        $this->db->escape( self::TABLE_NAME ),
        $this->db->escape( $fqdn )
      );

      if( [] === ( $result = $this->db->fetch( $sql ) ) )
      {
        Errors::set( 'Domain not found' );
        return null;
      }

      $result['active'] = Validate::toBool( $result['active'] );

      return new Domain( $this->db, $result );
    }

    /**
     * basic search function to list domain names based on a partial word
     *
     * @param string $query
     *      the string to try and match against the FQDN
     *
     * @return array
     *      array with matching domains
     */
    public function searchByPartialName( string $query ) : array
    {
      $sql = sprintf( '
        SELECT
          id, fqdn, ns, mbox, serial,
          refresh, retry, expire, minimum, ttl,
          active, xfer, owner_id
        FROM
          %s
        WHERE
          fqdn LIKE \'%%%s%%\'
        ORDER BY
          fqdn ASC
        ',
        $this->db->escape( self::TABLE_NAME ),
        $this->db->escape( $query )
      );

      $domains = [];

      foreach( $this->db->fetchAll( $sql ) as $domain )
      {
        $domain['active'] = Validate::toBool( $domain['active'] );
        $domains[$domain['id']] = new Domain( $this->db, $domain );
      }

      return $domains;
    }

    /**
     * get list of domains that belong to a specific owner
     * 
     * @param int $id
     *        the ID of the owner
     * 
     * @return array
     *        array with domains belonging to owner
     */
    public function getByOwnerId( int $id ) : array
    {
      $sql = sprintf( '
        SELECT
          id, fqdn, ns, mbox, serial,
          refresh, retry, expire, minimum, ttl,
          active, xfer, owner_id
        FROM
          %s
        WHERE
          owner_id = %d
        ORDER BY
          fqdn
        ',
        $this->db->escape( self::TABLE_NAME ),
        $id
      );

      $result = $this->db->fetchAll( $sql );

      $domains = [];

      foreach( $result as $domain )
      {
        $domain['active'] = Validate::toBool( $domain['active'] );
        $domains[$domain['id']] = new Domain( $this->db, $domain );
      }

      return $domains;
    }

    /**
     * get list of Master DNS servers from database
     * 
     * @param bool $all
     *        [optional] should we grab all or only active masters?
     * 
     * @return array
     *        array of master servers
     */
    public function getMasters( bool $all = false ) : array
    {
      $sql = sprintf( '
        SELECT
          id, fqdn, ip, active
        FROM
          %s
        %s
        ',
        $this->db->escape( self::MASTER_TABLE_NAME ),
        ( false === $all ) ? 'WHERE active = true' : ''
      );

      $result = $this->db->fetchAll( $sql );

      $masters = [];

      foreach( $result as $master )
      {
        $masters[] = new Master( $this->db, $master );
      }

      return $masters;
    }

    /**
     * create new instance of a domain to populate
     * 
     * @return Domain
     *        the new Domain
     */
    public function new() : Domain
    {
      return new Domain( $this->db, [] );
    }

    /**
     * export domain zone information into BIND format to the specified output folder
     * 
     * @param string $outputFolder
     *        the folder path to save domain zone data
     * 
     * @return bool
     *        returns if the export was successful or not
     */
    public function export( string $outputFolder ) : bool
    {
      // check permissions
      if( !is_writable( $outputFolder ) )
      {
        Errors::set( 'Unable to access output folder' );
        return false;
      }

      // get list of all domains
      if( [] === ( $domains = self::getAll() ) )
      {
        Errors::set( 'No domains to export' );
        return false;
      }

      foreach( $domains as $domain )
      {
        // skip inactive domains
        if( false === $domain->get( 'active' ) )
          continue;

        // export domain
        if( false === $domain->export( $outputFolder ) )
          return false;
      }

      return true;
    }

    /**
     * export either MASTER or SLAVE format of zone data
     * 
     * @param string $outputFolder
     *        the folder path to save domain zone data
     * 
     * @param int $type
     *        [optional] the type of export
     *        * Domains::EXPORT_ALL (default)
     *        * Domains::EXPORT_MASTER
     *        * Domains::EXPORT_SLAVE
     * 
     * @return bool
     *        returns if the export was successful or not
     */
    public function exportZones( string $outputFolder, int $type = self::EXPORT_ALL ) : bool
    {
      // if ALL is passed, simply run both master/slave
      if( self::EXPORT_ALL === $type )
      {
        if( false === self::exportZones( $outputFolder, self::EXPORT_MASTER ) )
          return false;

        if( false === self::exportZones( $outputFolder, self::EXPORT_SLAVE ) )
          return false;

        return true;
      }

      // check permissions
      if( !is_writable( $outputFolder ) )
      {
        Errors::set( 'Unable to access output folder' );
        return false;
      }

      // get list of all domains
      if( [] === ( $domains = self::getAll() ) )
      {
        Errors::set( 'No domains to export' );
        return false;
      }

      $zones = [];
      if( self::EXPORT_SLAVE === $type )
      {
        // grab list of master nameservers
        $masters = [];
        foreach( self::getMasters() as $master )
          $masters[] = $master->get( 'ip' );

        // bail if no masters are available
        if( empty( $masters ) )
          return false;

        $masters = implode( ";\n    ", $masters );

        // declare master nodes
        $zones[] = <<< MASTER
masters master_nodes {
\t{$masters};
};

MASTER;

      }

      foreach( $domains as $domain )
      {
        // skip inactive domains
        if( false === $domain->get( 'active' ) )
          continue;

        // export domain
        if( false === $domain->export( $outputFolder ) )
          return false;

        $zone = substr( $domain->get( 'fqdn' ), 0, -1 );

        switch( $type )
        {
          case self::EXPORT_MASTER:
            $slaves = implode( ";\n\t\t", explode( ',', $domain->get( 'xfer' ) ) );

            $zones[] = <<< DNS
zone "{$zone}" in {
\ttype master;
\tfile "{$zone}.zone";
\tnotify yes;
\talso-notify {
\t\t{$slaves};
\t};
};

DNS;

            break;
          case self::EXPORT_SLAVE:
            $zones[] = <<< DNS
zone "{$zone}" in {
\ttype slave;
\tfile "{$zone}.zone";
\tnotify no;
\tmasters { master_nodes; };
};

DNS;

            break;
        }
      }

      // generate slave and master zones
      $file = sprintf( '%s/zones.%s.list', $outputFolder, $type );
      file_put_contents( $file, implode( "\n", $zones ) );

      return true;
    }
  }
