<?php

  namespace Laodc\Dns;

  use Laodc\Database\Database;

  class Options
  {
    private Database $db;
    private array $data_fields = [];
    private ?int $domain_id = null;
    const TABLE_NAME = 'options';

    public function toArray() : array
    {
      return $this->data_fields;
    }

    function __construct( Database $database, ?int $domain_id = null )
    {
      $this->db = $database;
      $this->domain_id = $domain_id;

      // grab options from database and cache it
      $sql = sprintf( '
        SELECT
          *
        FROM
          %s
        WHERE
          domain_id %s
        ORDER BY
          option ASC
        ',
        self::TABLE_NAME,
        ( null === $domain_id ) ? 'is NULL' : sprintf( '= %d', $domain_id )
      );

      foreach( $this->db->fetchAll( $sql ) as $option )
      {
        $this->data_fields[$option['option']] = $option['value'];
      }
    }

    /**
     * @param string $key
     *        the key name of the option to assign
     *
     * @param string $value
     *        the value to store on the database
     * 
     * @return bool
     *        true if added to database
     */
    public function set( string $key, string $value ) : bool
    {
      $key = strtolower( $key );

      $sql = sprintf( '
        INSERT INTO
          %s
        ( domain_id, option, value )
        VALUES
        ( %s, \'%s\', \'%s\' )
        ON CONFLICT( option, domain_id ) DO UPDATE
          SET
            domain_id = EXCLUDED.domain_id,
            option = EXCLUDED.option,
            value = EXCLUDED.value
        RETURNING
          option, value
        ',
        self::TABLE_NAME,
        $this->domain_id ?: 'NULL',
        $this->db->escape( $key ),
        $this->db->escape( $value )
      );

      if( false === ( $result = $this->db->fetch( $sql ) ) )
        return false;

      // update cache
      $this->data_fields[$result['option']] = $result['value'];

      return true;
    }

    /**
     * @param string $key
     *        the key name of the option to grab
     *
     * @param bool $force
     *        [optional] force the data from database
     *
     * @return string|null
     *        returns the option value otherwise NULL if doesn't exist
     */
    public function get( string $key, bool $force = false ) :? string
    {
      $key = strtolower( $key );

      if( true === $force || !isset( $this->data_fields[$key] ) )
      {
        $sql = sprintf( '
          SELECT
            value
          FROM
            %s
          WHERE
            domain_id %s AND
            option LIKE \'%s\'
          ',
          self::TABLE_NAME,
          ( null === $this->domain_id ) ? 'is NULL' : sprintf( '= %d', $this->domain_id ),
          $this->db->escape( $key )
        );

        if( [] === ( $result = $this->db->fetch( $sql ) ) )
          return null;

        // update cache
        $this->data_fields[$key] = $result['value'];
      }

      return $this->data_fields[$key] ?: null;
    }

    /**
     * removed an option from the database
     *
     * @param string $key
     *        the key name of the option to remove
     * 
     * @return bool
     *        true if removed from the database
     */
    public function del( string $key ) : bool
    {
      $key = strtolower( $key );

      $sql = sprintf( '
        DELETE FROM
          %s
        WHERE
          domain_id %s AND
          option LIKE \'%s\'
        ',
        self::TABLE_NAME,
        ( null === $this->domain_id ) ? 'is NULL' : sprintf( '= %d', $this->domain_id ),
        $this->db->escape( $key )
      );

      if( false === $this->db->query( $sql ) )
        return false;

      unset( $this->data_fields[$key] );
      return true;
    }

    /**
     * delete all option settings from database
     * 
     * @return bool
     *        true if successfully removed all options from database
     */
    public function del_all() : bool
    {
      $sql = sprintf( '
        DELETE FROM
          %s
        WHERE
          domain_id %s
        ',
        self::TABLE_NAME,
        ( null === $this->domain_id ) ? 'is NULL' : sprintf( '= %d', $this->domain_id )
      );

      if( false === $this->db->query( $sql ) )
        return false;

      unset( $this->data_fields );
      return true;
    }
  }
