<?php

  namespace Laodc\Dns;

  use Laodc\Database\Database;
  use Laodc\Errors\Errors;

  class Record extends Base
  {
    private Database $db;
    private Domain $domain;

    function __construct( Database $database, Domain $domain, array $record )
    {
      $this->db = $database;
      $this->domain = $domain;
      $this->populate( [ Validate::class, 'formatRecordData' ], $record );
    }

    public static function getTypes() : array
    {
      return array_column( RecordTypes::cases(), 'value' );
    }

    public function save( bool $updateSerial = false ) : bool
    {
      if( empty( $this->domain ) || empty( $this->domain->get( 'id' ) ) )
      {
        Errors::set( 'Record is not linked to a domain' );
        return false;
      }

      // allowed fields
      $fields = [
        'name'    => false,
        'type'    => true,
        'data'    => true,
        'pref'    => false,
        'ttl'     => false,
        'active'  => false,
      ];

      foreach( $fields as $field => $required )
      {
        $value = $this->get( $field );
        if( true === $required && empty( $value ) && 0 === strlen( $value ) )
        {
          Errors::set( sprintf( '%s is required', $field ) );
          return false;
        }

        // sanatize record data
        $this->set( $field, Validate::formatRecordData( $field, $this->get( $field ) ) );
      }

      // check type if valid type
      if( false === in_array( $this->get( 'type' ), static::getTypes() ) )
      {
        Errors::set( sprintf( '%s isn\'t a valid record type', $this->get( 'type' ) ) );
        return false;
      }

      // Set name to FQDN is blank or @ was submitted
      if( empty( $this->get( 'name' ) ) || '@' === $this->get( 'name' ) )
        $this->set( 'name', $this->domain->get( 'fqdn' ) );

      // make sure name is valid
      if( null === ( $name = Validate::Name( $this->get( 'name' ), $this->domain->get( 'fqdn' ), true ) ) )
        return false;

      $this->set( 'name', $name );

      // check that ttl is correct foramt
      $this->set( 'ttl', $this->get( 'ttl' ) ?? Records::DEFAULT_TTL );

      if( false === Validate::isInt32( $this->get( 'ttl' ), Records::MIN_TTL ) )
        return false;

      // verify ttl is not lower than domain ttl
      if( $this->get( 'ttl' ) < $this->domain->get( 'minimum' ) )
      {
        Errors::set( 'ttl is less than the domain ttl' );
        return false;
      }

      // make sure pref is a number in range
      if( $this->get( 'pref' ) && false === Validate::inRange( $this->get( 'pref' ) ) )
      {
        Errors::set( 'pref must be a number between 0 and 65535' );
        return false;
      }

      // Validate against record type
      switch( $this->get( 'type' ) )
      {
        case RecordTypes::A->value:
          if( false === Validate::IPv4( $this->get( 'data' ) ) )
            return false;
          break;
        case RecordTypes::AAAA->value:
          if( false === Validate::IPv6( $this->get( 'data' ) ) )
            return false;
          break;
        case RecordTypes::CAA->value:
          if( null === ( $data = Validate::CAA( $this->get( 'data' ) ) ) )
            return false;

          $this->set( 'data', $data );
          break;
        case RecordTypes::ALIAS->value:
        case RecordTypes::CH->value:
        case RecordTypes::CNAME->value:
        case RecordTypes::MX->value:
        case RecordTypes::NS->value:
          if( null === ( $data = Validate::Name( $this->get( 'data' ), $this->domain->get( 'fqdn' ) ) ) )
            return false;

          $this->set( 'data', $data );
          break;
        case RecordTypes::PTR->value:
          if( null === ( $data = Validate::Name( $this->get( 'data' ), $this->domain->get( 'fqdn' ), false ) ) )
            return false;

          if( '.' !== substr( $data, -1 ) )
          {
            Errors::set( 'data field must end in a dot (.)' );
            return false;
          }

          $this->set( 'data', $data );
          break;
        case RecordTypes::NAPTR->value:
          if( null === ( $data = Validate::NAPTR( $this->get( 'data' ) ) ) )
            return false;

          $this->set( 'data', $data );
          break;
        case RecordTypes::RP->value:
          if( null === ( $data = Validate::RP( $this->get( 'data' ), $this->domain->get( 'fqdn' ) ) ) )
            return false;

          $this->set( 'data', $data );
          break;
        case RecordTypes::SRV->value:
          if( null === ( $data = Validate::SRV( $this->get( 'data' ), $this->domain->get( 'fqdn' ) ) ) )
            return false;

          $this->set( 'data', $data );
          break;
        case RecordTypes::TLSA->value:
          if( null === ( $data = Validate::TLSA( $this->get( 'data' ), $this->domain->get( 'fqdn' ) ) ) )
            return false;

          $this->set( 'data', $data );
          break;
        case RecordTypes::TXT->value:
          if( false === Validate::TXT( $this->get( 'data' ) ) )
            return false;
          break;
        case RecordTypes::URL->value:
        case RecordTypes::URL301->value:
          if( false === parse_url( $this->get( 'data' ) ) )
            return false;
          break;
        default:
          Errors::set( 'invalid record type' );
          return false;
      }

      // check for existing record
      // if ID is specified, don't match against itself
      $sql = sprintf( '
        SELECT
          id
        FROM
          %s
        WHERE
          %s
          domain_id = %d AND
          name = \'%s\' AND
          type = \'%s\' AND
          data = \'%s\'
        ',
        $this->db->escape( Records::TABLE_NAME ),
        ( $this->get( 'id' ) ) ? sprintf( 'id <> %d AND', $this->get( 'id' ) ) : '',
        $this->domain->get( 'id' ),
        $this->db->escape( $this->get( 'name' ) ),
        $this->db->escape( $this->get( 'type' ) ),
        $this->db->escape( $this->get( 'data' ) )
      );

      // record exists, bail out
      if( [] !== $this->db->fetch( $sql ) )
      {
        Errors::set( 'Existing record found' );
        return false;
      }

      // add new record
      $sql = sprintf( '
        INSERT INTO
          %s
        (
          domain_id, name, type,
          data, pref, ttl, active
        )
        VALUES
          ( %d, \'%s\', \'%s\', \'%s\', %d, %d, %s )
        RETURNING
          id, domain_id, name, type,
          data, pref, ttl, active
        ',
        $this->db->escape( Records::TABLE_NAME ),
        $this->domain->get( 'id' ),
        $this->db->escape( $this->get( 'name' ) ),
        $this->db->escape( $this->get( 'type' ) ),
        $this->db->escape( $this->get( 'data' ) ),
        $this->get( 'pref' ),
        $this->get( 'ttl' ),
        Validate::boolToSQL( $this->get( 'active' ) )
      );

      // update record if ID exists
      if( !empty( $this->get( 'id' ) ) )
      {
        $sql = sprintf( '
          UPDATE
            %s
          SET
            name = \'%s\',
            type = \'%s\',
            data = \'%s\',
            pref = %d,
            ttl = %d,
            active = %s
          WHERE
            domain_id = %d AND
            id = %d
          RETURNING
            id, domain_id, name, type,
            data, pref, ttl, active
          ',
          $this->db->escape( Records::TABLE_NAME ),
          $this->db->escape( $this->get( 'name' ) ),
          $this->db->escape( $this->get( 'type' ) ),
          $this->db->escape( $this->get( 'data' ) ),
          $this->get( 'pref' ),
          $this->get( 'ttl' ),
          Validate::boolToSQL( $this->get( 'active' ) ),
          $this->domain->get( 'id' ),
          $this->get( 'id' )
        );
      }

      if( [] === ( $record = $this->db->fetch( $sql ) ) )
      {
        Errors::set( 'Unable to add record to database' );
        return false;
      }

      $record['active'] = Validate::toBool( $record['active'] );

      foreach( $record as $field => $value )
        $this->set( $field, $value );

      if( true === $updateSerial )
        $this->domain->updateSerial();

      return true;
    }

    public function del() : bool
    {
      if( empty( $this->get( 'id' ) ) )
      {
        Errors::set( 'Record does not exist' );
        return false;
      }

      // delete record
      $sql = sprintf( '
        DELETE FROM
          %s
        WHERE
          id = %d
        ',
        Records::TABLE_NAME,
        $this->get( 'id' )
      );

      if( [] === $this->db->fetch( $sql ) )
      {
        Errors::set( 'Error deleting record' );
        return false;
      }

      // reset fields
      $this->data_fields = [];
      return true;
    }
  }
