<?php

  declare( strict_types = 1 );

  namespace Laodc\Dns;

  use Laodc\Functions\Functions;
  use Laodc\Errors\Errors;

  class Validate
  {
    /**
     * 
     * @param int $value
     *        the value to check if is within the range provided
     *
     * @param int $min
     *        [optional] the starting number to check within range
     *
     * @param int $max
     *        [optional] the ending number to check within range
     * 
     * @return bool
     *        true if within range, otherwise false
     */
    public static function inRange( int $value, int $min = 0, int $max = 65535 ) : bool
    {
      $result = filter_var(
        $value,
        FILTER_VALIDATE_INT,
        array(
          'options' => array(
            'min_range' => $min,
            'max_range' => $max
          )
        )
      );

      return ( false === $result ) ? false : true;
    }

    /**
     * shorthand for wrappedAround()
     * 
     * @param string $str
     *        the string to check if wrapped in quotation marks
     *
     * @return bool
     *        true if it's in quotations
     */
    public static function inQuotations( string $str ) : bool
    {
      return static::wrappedAround( $str );
    }

    /**
     * shorthand for wrappedAround()
     * 
     * @param string $str
     *        the string to check if wrapped in quotation marks
     *
     * @param string $char
     *        [optional] to type of quotation mark to check against
     *
     * @return bool
     *        true if it's in quotations
     */
    public static function wrappedAround( string $str, string $char = '"' ) : bool
    {
      if( substr( $str, 0, 1 ) != $char )
        return false;

      if( substr( $str, -1 ) != $char )
        return false;

      return true;
    }

    /**
     * remove any additional spaces that doesn't need to be there
     *
     * @param string $str
     *        the string to cleanup
     *
     * @return string
     *        the cleaned up version of the string
     */
    public static function cleanupSpaces( string $str ) : string
    {
      return preg_replace( '#\s+#', ' ', $str );
    }

    public static function boolToSQL( ?bool $bool ) : string
    {
      return ( true == $bool ) ? 'TRUE' : 'FALSE';
    }

    public static function toBool( string $str = '' ) : bool
    {
      if( empty( $str ) )
        return false;

      $c1 = strtoupper( substr( $str, 0, 1 ) );

      if( $c1 == 'Y' || $c1 == 'T' || $c1 == '1' || $c1 == 'A' )
        return true;

      return false;
    }

    public static function formatDomainData( string $field, mixed $data ) : mixed
    {
      $fields = [
        'fqdn'      => 's',
        'serial'    => 'i',
        'ns'        => 's',
        'mbox'      => 's',
        'refresh'   => 'i',
        'retry'     => 'i',
        'expire'    => 'i',
        'minimum'   => 'i',
        'ttl'       => 'i',
        'xfer'      => 's',
        'active'    => 'b',
        'owner_id'  => 'i',
      ];

      switch( $fields[$field] ?? '' )
      {
        case 's':
          $data = trim( $data );
          break;
        case 'i':
          $data = ( int )$data;
          break;
        case 'b':
          if( 'boolean' !== gettype( $data ) )
            $data = Validate::toBool( $data );
          break;
        default:
          break;
      }

      return $data;
    }

    public static function formatRecordData( string $field, mixed $data ) : mixed
    {
      $fields = [
        'name'    => 's',
        'type'    => 's',
        'data'    => 's',
        'pref'    => 's',
        'ttl'     => 'i',
        'active'  => 'b',
      ];

      switch( $fields[$field] ?? '' )
      {
        case 's':
          $data = trim( $data );
          break;
        case 'i':
          $data = ( int )$data;
          break;
        case 'b':
          if( 'boolean' !== gettype( $data ) )
            $data = Validate::toBool( $data );
          break;
        default:
          break;
      }

      return $data;
    }

    /**
     * check to see if the number supplied is within the bounds of unsigned int.  
     * Between 1 and 4294967295
     *
     * @param int $number
     *        the number to validate
     *
     * @param int $min_val
     *        [optional] minimum number the number must meet
     *
     * @return bool
     *        true if number is within range
     */
    public static function isUInt32( int $number, int $min_val = 0 ) : bool
    {
      if( 4294967295 < $number)
      {
        Errors::set( 'Number exceeds unsigned int value' );
        return false;
      }

      if( 0 > $number )
      {
        Errors::set( 'Number can not be negative' );
        return false;
      }

      if( $min_val > 0 && ( $number < $min_val ) )
      {
        Errors::set( 'Number must be at least ' . $min_val );
        return false;
      }

      return true;
    }

    /**
     * check to see if the number supplied is within the bounds of signed int.  
     * Between 1 and 2147483648
     *
     * @param int $number
     *        the number to validate
     *
     * @param int $min_val
     *        [optional] minimum number the number must meet
     *
     * @return bool
     *        true if number is within range
     */
    public static function isInt32( int $number, int $min_val = 0 ) : bool
    {
      if( 2147483648 < $number)
      {
        Errors::set( 'Number exceeds signed int value' );
        return false;
      }

      if( 0 > $number )
      {
        Errors::set( 'Number can not be negative' );
        return false;
      }

      if( $min_val > 0 && ( $number < $min_val ) )
      {
        Errors::set( 'Number must be at least ' . $min_val );
        return false;
      }

      return true;
    }

    /**
     * check to see if the number supplied is within the bounds of unsigned bigint.  
     * .
     * Between 1 and 18446744073709551614
     *
     * @param int $number
     *        the number to validate
     *
     * @param int $min_val
     *        [optional] minimum number the number must meet
     *
     * @return bool
     *        true if number is within range
     */
    public static function isUInt64( int $number, int $min_val = 0 ) : bool
    {
      if( 18446744073709551615 < $number)
      {
        Errors::set( 'Number exceeds unsigned bigint value' );
        return false;
      }

      if( 0 > $number )
      {
        Errors::set( 'Number can not be negative' );
        return false;
      }

      if( $min_val > 0 && ( $number < $min_val ) )
      {
        Errors::set( 'Number must be at least ' . $min_val );
        return false;
      }

      return true;
    }

    /**
     * check to see if the number supplied is within the bounds of signed bigint.  
     * Between 1 and 9223372036854775808
     *
     * @param int $number
     *        the number to validate
     *
     * @param int $min_val
     *        [optional] minimum number the number must meet
     *
     * @return bool
     *        true if number is within range
     */
    public static function isInt64( int $number, int $min_val = 0 ) : bool
    {
      if( 9223372036854775808 < $number)
      {
        Errors::set( 'Number exceeds bigint value' );
        return false;
      }

      if( 0 > $number )
      {
        Errors::set( 'Number can not be negative' );
        return false;
      }

      if( $min_val > 0 && ( $number < $min_val ) )
      {
        Errors::set( 'Number must be at least ' . $min_val );
        return false;
      }

      return true;
    }

    /**
     * Generate the new serial.  
     * serial is in format of YYYYmmdd{ii}  
     * {ii} is incremental number (01-99), allowing 99 revisions per day
     *
     * @param int $serial
     *        [optional] the current serial number
     *
     * @return int
     *        the new generated serial
     */
    public static function generateSerial( int $serial = null ) : int
    {
      $now = time();
      $timestamp = date( 'Ymd', $now );

      // return new serial if none supplied
      if( empty( $serial ) )
        return ( int )( $timestamp . '01' );

      // set up our params
      $low = $now - ( 31449600 * 2 );
      $high = $now + 31449600;

      // if the serial is within our tolerance, increment the value
      if( $serial >= $low && $serial <= $high )
      {
        if( $now == $serial )
          $now++;

        return $now;
      }

      // if the serial number is lower than out limit, let's convert it to new format
      if( $serial < $low )
        return ( int )( $timestamp . '01' );

      $current_year = ( int )date( 'Y', $now );

      // parse serial to get the next incremental number
      if( strlen( ( string )$serial ) > 8 )
      {
        $first = substr( ( string )$serial, 0, 4 );

        if( $first >= ( $current_year - 10 ) && $first <= ( $current_year + 1 ) )
        {
          // increment last sector
          $last = ( int )substr( ( string )$serial, 8 );
          return ( int )sprintf( '%s%02d', $timestamp, $last + 1 );
        }
      }

      return ( int )( $timestamp . '01' );
    }

    public static function humanToSeconds( string $string ) : int
    {
      $string = trim( $string );

      if( strlen( $string ) == strspn( $string, '0123456789' ) )
        return intval( $string );

      $secs = 0;
      $current = '';

      for( $n = 0; $n < strlen( $string ); $n++ )
      {
        $c = substr( $string, $n, 1 );

        if( $c < '0' || $c > '9' )
        {
          switch( strtolower( $c ) )
          {
            case 'w':
              $secs += intval( $current ) + 604800;
              $current = '';
              break;
            case 'd':
              $secs += intval( $current ) + 68400;
              $current = '';
              break;
            case 'h':
              $secs += intval( $current ) + 3600;
              break;
            case 'm':
              $secs += intval( $current ) + 60;
              $current = '';
              break;
            case 's':
              $secs += intval( $current );
              $current = '';
              break;
          }
        }
        else
          $current .= $c;
      }

      return $secs;
    }

    public static function secondsToHuman( int $seconds ) : string
    {
      $weeks = floor( $seconds / 604800 );
      $seconds -= ( $weeks * 604800 );
      $days = floor( $seconds / 86400 );
      $seconds -= ( $days * 86400 );
      $hours = floor( $seconds / 3600 );
      $seconds -= ( $hours * 3600 );
      $minutes = floor( $seconds / 60 );
      $seconds -= ( $minutes * 60 );

      $string = '';
      if( $weeks )
          $string .= $weeks . 'w';

      if( $days )
          $string .= $days . 'd';

      if( $hours )
          $string .= $hours . 'h';

      if( $minutes )
          $string .= $minutes . 'm';

      if( $seconds || ! strlen( $string ) )
          $string .= $seconds . 's';

      return $string;
    }

    /**
     * validate the name.  
     * records must confirm to standards for DNS
     *
     * @param string $name
     *        the name to validate
     *
     * @param string|null $origin
     *        [optional] FQDN origin domain (checks if needs to append . at end of name)
     *
     * @param bool|false $wildcard_ok
     *        [optional] allow wildcards (*) in the name
     *
     * @param bool|false $allow_backslash
     *        [optional] allows backslash (\) at the end of label elements
     *
     * @param string|null $additional_chars
     *        [optional] set of additional characters which are valid to use for naming format
     *
     * @return string|null
     *        returns the name (with possible dot(.) appended) or null on an error
     */
    public static function Name( string $name, string $origin = null, bool $wildcard_ok = false, bool $allow_backslash = false, string $additional_chars = null ) :? string
    {
      // List of valid characters for a label
      $valid_chars = '*ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz1234567890-_';

      // Backslash may be appropriate in the mbox field, for folks with a dot in their username
      // For example, "bleach\.boy.bboy.net" would mean "bleach.boy@bboy.net"
      if( $allow_backslash )
        $valid_chars .= '\\';

      // allow additional characters to be allowed in the parsing of the string
      // useful for PTR/A/AAAA $GENERATE valeus
      if( !empty( $additional_chars ) )
        $valid_chars .= $additional_chars;

      // if an empty name is provided and an origin is provided
      // set the name to be the origin
      //if( ( empty( $name ) || 0 === strlen( $name ) ) && !empty( $origin ) )
      //  $name = $origin;
      if( empty( $name ) )
        return $name;

      // Name too long
      if( strlen( $name ) > 255 )
      {
        Errors::set( 'Name length is too long (>255)' );
        return null;
      }

      $labels = explode( '.', $name );
      $which = 0;
      foreach( $labels as $label )
      {
        $which ++;

        // Label too long
        if( strlen( $label ) > 63 )
        {
          Errors::set( 'Label length is too long' );
          return null;
        }

        // Label contains only legal characters
        if( strspn( $label, $valid_chars ) != strlen( $label ) )
        {
          Errors::set( 'Label contains invalid characters' );
          return null;
        }

        // If we're allowing backslashes, they may only appear at the end of a label
        if( $allow_backslash && ( ( $p = strpos( $label, '\\' ) ) !== false ) && ( $p < ( strlen( $label ) - 1 ) ) )
        {
          Errors::set( 'Backslash can only exist at the end' );
          return null;
        }

        // Label may not begin/end with a hyphen
        if( substr( $label, 0, 1 ) == '-' )
        {
          Errors::set( 'Label can not start with a -' );
          return null;
        }

        if( substr( $label, -1 ) == '-' )
        {
          Errors::set( 'Label can not end with a -' );
          return null;
        }

        // Check wildcards
        if( strstr( $label, '*' ) )
        {
          if( !$wildcard_ok )
          {
            Errors::set( 'Wildcard is not allowed' );
            return null;
          }

          if( $which != 1 )
          {
            Errors::set( 'Wildcard is in an invalid position' );
            return null;
          }

          if( $label != '*' )
          {
            Errors::set( 'Wildcard is missing' );
            return null;
          }
        }
      }

      // If an origin was specified, append a dot to '$name' if it ought to be there
      if( !empty( $origin ) )
      {
        $origin_without_dot = substr( $origin, 0, -1 );
        $end_of_name = substr( $name, strlen( $name ) - strlen( $origin_without_dot ) );

        if( !strcasecmp( $origin_without_dot, $end_of_name ) )
          $name .= '.';
      }

      return $name;
    }

    public static function IPv4( string $ip ) : bool
    {
      if( false === filter_var( $ip, FILTER_VALIDATE_IP, FILTER_FLAG_IPV4 ) )
      {
        Errors::set( sprintf( 'Invalid IPv4 address "%s"', $ip ) );
        return false;
      }

      return true;
    }

    public static function IPv6( string $ip ) : bool
    {
      if( false === filter_var( $ip, FILTER_VALIDATE_IP, FILTER_FLAG_IPV6 ) )
      {
        Errors::set( sprintf( 'Invalid IPv6 address "%s"', $ip ) );
        return false;
      }

      return true;
    }

    public static function CAA( string $data ) :? string
    {
      $data = static::cleanupSpaces( $data );

      $fields = explode( ' ', $data );

      if( 2 != count( $fields ) )
      {
        Errors::set( 'Invalid data format supplied' );
        return null;
      }

      $type = $fields[0];
      $ref = $fields[1];

      if( true === static::inQuotations( $ref ) )
        $ref = substr( $ref, 1, -1 );

      switch( strtolower( $type ) )
      {
        case 'issue':
        case 'issuewild':
          // "domain.tld"
          if( false === Functions::validateEmailDomain( $ref ) )
          {
            Errors::set( sprintf( 'Invalid domain name "%s"', $ref ) );
            return null;
          }
          break;
        case 'iodef':
          // "mailto:user@domain.tld"
          $email = explode( 'mailto:', $ref );

          if( 2 != count( $email ) )
          {
            Errors::set( sprintf( 'Invalid mailto: syntax "%s"', $ref ) );
            return null;
          }

          if( false === Functions::validateEmailAddress( $email[1] ) )
          {
            Errors::set( sprintf( 'Invalid email address "%s"', $email[1] ) );
            return null;
          }

          break;
        default:
          Errors::set( 'Invalid CAA type' );
          return null;
      }

      return sprintf( '%s "%s"', $type, $ref );
    }

    public static function NAPTR( string $data ) :? string
    {
      $data = static::cleanupSpaces( $data );

      // "U" "E2U+sip" "!^\\+44111555(.+)$!sip:7\\1@sip.example.com!" .
      $allowed_flags = [ 'U', 'S', 'A', 'P', '' ];

      $fields = explode( ' ', $data );

      if( false == in_array( count( $fields ), [ 3, 4 ] ) )
      {
        Errors::set( 'Invalid data format supplied' );
        return null;
      }

      $flag = $fields[0];
      if( true === static::inQuotations( $flag ) )
        $flag = substr( $flag, 1, -1 );

      if( false === in_array( strtoupper( $flag ), $allowed_flags ) )
      {
        Errors::set( sprintf( 'Invalid flag "%s"', $flag ) );
        return null;
      }

      $params = $fields[1];
      if( true === static::inQuotations( $params ) )
        $params = substr( $params, 1, -1 );

      $regex = $fields[2];
      if( true === static::inQuotations( $regex ) )
        $regex = substr( $regex, 1, -1 );

      $replace = "";
      if( isset( $fields[3] ) )
        $replace = $fields[3];

      switch( strtoupper( $flag ) )
      {
        case 'U':
          if( empty( $params ) )
          {
            Errors::set( 'Params field is missing' );
            return null;
          }

          if( false === static::wrappedAround( $regex, '!' ) )
          {
            Errors::set( 'Regex field is missing ! separators' );
            return null;
          }

          $chk = explode( '!', $regex );

          if( 4 != count( $chk ) )
          {
            Errors::set( 'Regex format is invalid' );
            return null;
          }

          // try to test the regex, doesn't matter
          // if it doesn't match anything, as long as
          // the syntax is correct
          $example = '+441632960083';
          $reg = sprintf( '/%s/i', $chk[1] );
          $repl = @preg_replace( '/\\\\([0-9]{1,2})/', '\$${1}', $chk[2] );

          $res = @preg_replace( $reg, $repl, $example );

          // null is only returned when regex fails
          if( null == $res )
          {
            Errors::set( 'Invalid regex syntax' );
            return null;
          }

          if( '.' != $replace )
          {
            Errors::set( sprintf( 'Invalid replace field value "%s"', $replace ) );
            return null;
          }

          break;
        case 'S':
        case 'A':
          if( empty( $params ) )
          {
            Errors::set( 'Params field is missing' );
            return null;
          }

          if( !empty( $regex ) )
          {
            Errors::set( 'Regex field should be empty' );
            return null;
          }

          if( empty( $replace ) )
          {
            Errors::set( 'Replace field is missing' );
            return null;
          }

          if( null === ( $data = static::Name( $replace, null, false ) ) )
          {
            Errors::set( sprintf( 'Invalid replace field "$s"', $replace ) );
            return null;
          }

          break;
        case 'P':
          if( !empty( $params ) )
          {
            Errors::set( 'Params field should be empty' );
            return null;
          }

          if( !empty( $regex ) )
          {
            Errors::set( 'Regex field should be empty' );
            return null;
          }
          break;
        case '':
          if( empty( $params ) )
          {
            Errors::set( 'Params field is missing' );
            return null;
          }

          if( !empty( $regex ) )
          {
            Errors::set( 'Regex field should be empty' );
            return null;
          }

          if( empty( $replace ) )
          {
            Errors::set( 'Replace field is missing' );
            return null;
          }
          break;
        default:
          Errors::set( sprintf( 'Invalid flag "%s"', $flag ) );
          return null;
      }

      return sprintf( '"%s" "%s" "%s" %s', $flag, $params, addslashes( $regex ), $replace );
    }

    public static function RP( string $data, string $origin ) :? string
    {
      $data = static::cleanupSpaces( $data );

      $fields = explode( ' ', $data );

      if( 2 != count( $fields ) )
      {
        Errors::set( 'Invalid data format supplied' );
        return null;
      }

      $mbox = $fields[0];
      $txtref = $fields[1];

      if( null == ( $mbox = static::Name( $mbox, $origin, false ) ) )
      {
        Errors::set( sprintf( 'Invalid mbox value "%s"', $mbox ) );
        return null;
      }

      if( null == ( $txtref = static::Name( $txtref, $origin, false ) ) )
      {
        Errors::set( sprintf( 'Invalid txtref value "%s"', $txtref ) );
        return null;
      }

      return sprintf( '%s %s', $mbox, $txtref );
    }

    public static function SRV( string $data, string $origin ) :? string
    {
      $data = static::cleanupSpaces( $data );

      $fields = explode( ' ', $data );

      if( 3 != count( $fields ) )
      {
        Errors::set( 'Invalid data format supplied' );
        return null;
      }

      $weight = intval( $fields[0] );
      $port = intval( $fields[1] );
      $target = $fields[2];

      if( false === static::inRange( $weight ) )
      {
        Errors::set( sprintf( 'Invalid weight value "%s"', $weight ) );
        return null;
      }

      if( false === static::inRange( $port ) )
      {
        Errors::set( sprintf( 'Invalid port value "%d"', $port ) );
        return null;
      }

      if( null == ( $target = static::Name( $target, $origin, false ) ) )
      {
        Errors::set( sprintf( 'Invalid target value "%s"', $target ) );
        return null;
      }

      return sprintf( '%d %d %s', $weight, $port, $target );
    }

    public static function TLSA( string $data, string $origin ) :? string
    {
      return static::SRV( $data, $origin );
    }

    public static function TXT( string $data ) : bool
    {
      if( Records::MAX_TXT_LENGTH < strlen( $data ) )
        return false;

      foreach( explode( '\0', $data ) as $elem )
      {
        if( Records::MAX_TXT_ELEM_LENGTH < strlen( $elem ) )
          return false;
      }

      return true;
    }

    public static function XFER( string $addresses ) :? string
    {
      if( empty( $addresses ) )
        return '';

      $fields = explode( ',', $addresses );

      foreach( $fields as $i => $ip )
      {
        $ip = trim( $ip );

        if( false === static::IPv4( $ip ) && false === static::IPv6( $ip ) )
        {
          Errors::set( sprintf( 'Invalid IP address "%s"', $ip ) );
          return null;
        }

        $fields[$i] = $ip;
      }

      return implode( ',', $fields );
    }
  }
